import pytest
import tensorflow.keras.backend as K
from tensorflow.keras.layers import Activation, Flatten, Reshape
from tensorflow.python.keras.backend import _get_available_gpus

from decomon.backward_layers.convert import to_backward


def test_Backward_NativeActivation_1D_box_model(n, activation, mode, floatx, decimal, helpers):
    dc_decomp = False

    #  tensor inputs
    inputs = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)

    # numpy inputs
    inputs_ = helpers.get_standard_values_1d_box(n, dc_decomp=dc_decomp)

    # keras layer
    keras_layer = Activation(activation, dtype=K.floatx())

    # get backward layer
    backward_layer = to_backward(keras_layer, mode=mode)

    # backward outputs
    outputs = backward_layer(inputs_for_mode)
    f_decomon = K.function(inputs, outputs)
    outputs_ = f_decomon(inputs_)

    # check bounds consistency
    helpers.assert_backward_layer_output_properties_box_linear(
        full_inputs=inputs_,
        backward_outputs=outputs_,
        decimal=decimal,
    )


def test_Backward_NativeActivation_multiD_box(odd, activation, floatx, decimal, mode, helpers):
    dc_decomp = False

    #  tensor inputs
    inputs = helpers.get_tensor_decomposition_multid_box(odd=odd, dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)

    # numpy inputs
    inputs_ = helpers.get_standard_values_multid_box(odd, dc_decomp=dc_decomp)

    # keras layer
    keras_layer = Activation(activation, dtype=K.floatx())

    # get backward layer
    backward_layer = to_backward(keras_layer, mode=mode)

    # backward outputs
    outputs = backward_layer(inputs_for_mode)
    f_decomon = K.function(inputs, outputs)
    outputs_ = f_decomon(inputs_)

    # check bounds consistency
    helpers.assert_backward_layer_output_properties_box_linear(
        full_inputs=inputs_,
        backward_outputs=outputs_,
        decimal=decimal,
    )


def test_Backward_NativeFlatten_multiD_box(odd, floatx, decimal, mode, data_format, helpers):
    if data_format == "channels_first" and not len(_get_available_gpus()):
        pytest.skip("data format 'channels first' is possible only in GPU mode")

    dc_decomp = False

    #  tensor inputs
    inputs = helpers.get_tensor_decomposition_multid_box(odd=odd, dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)

    # numpy inputs
    inputs_ = helpers.get_standard_values_multid_box(odd, dc_decomp=dc_decomp)

    # keras layer
    keras_layer = Flatten(data_format, dtype=K.floatx())

    # get backward layer
    backward_layer = to_backward(keras_layer, mode=mode)

    # backward outputs
    outputs = backward_layer(inputs_for_mode)
    f_decomon = K.function(inputs, outputs)
    outputs_ = f_decomon(inputs_)

    # check bounds consistency
    helpers.assert_backward_layer_output_properties_box_linear(
        full_inputs=inputs_,
        backward_outputs=outputs_,
        decimal=decimal,
    )


def test_Backward_NativeReshape_multiD_box(odd, floatx, decimal, mode, helpers):
    dc_decomp = False

    #  tensor inputs
    inputs = helpers.get_tensor_decomposition_multid_box(odd=odd, dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)

    # numpy inputs
    inputs_ = helpers.get_standard_values_multid_box(odd, dc_decomp=dc_decomp)

    # keras layer
    keras_layer = Reshape((-1,), dtype=K.floatx())

    # get backward layer
    backward_layer = to_backward(keras_layer, mode=mode)

    # backward outputs
    outputs = backward_layer(inputs_for_mode)
    f_decomon = K.function(inputs, outputs)
    outputs_ = f_decomon(inputs_)

    # check bounds consistency
    helpers.assert_backward_layer_output_properties_box_linear(
        full_inputs=inputs_,
        backward_outputs=outputs_,
        decimal=decimal,
    )



import warnings
from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.types.experimental import TensorLike

from decomon.core import (
    BoxDomain,
    ForwardMode,
    InputsOutputsSpec,
    PerturbationDomain,
    Slope,
    get_affine,
    get_ibp,
)
from decomon.layers.core import DecomonLayer
from decomon.layers.utils import exp, expand_dims, frac_pos, multiply, softplus_, sum
from decomon.utils import (
    get_linear_hull_s_shape,
    minus,
    relu_,
    sigmoid_prime,
    softsign_prime,
    tanh_prime,
)

ELU = "elu"
SELU = "selu"
SOFTPLUS = "softplus"
SOFTSIGN = "softsign"
SOFTMAX = "softmax"
RELU = "relu"
SIGMOID = "sigmoid"
TANH = "tanh"
EXPONENTIAL = "exponential"
HARD_SIGMOID = "hard_sigmoid"
LINEAR = "linear"
GROUP_SORT_2 = "GroupSort2"


def relu(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    alpha: float = 0.0,
    max_value: Optional[float] = None,
    threshold: float = 0.0,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """
    Args:
        inputs: list of input tensors
        dc_decomp: boolean that indicates
        perturbation_domain: type of convex input domain (None or dict)
        alpha: see Keras official documentation
        max_value: see Keras official documentation
        threshold: see Keras official documentation
        mode: type of Forward propagation (ibp, affine, or hybrid)
        slope:
        **kwargs: see Keras official documentation
    whether we return a difference of convex decomposition of our layer

    Returns:
        the updated list of tensors
    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    if threshold != 0:
        raise NotImplementedError()

    if not alpha and max_value is None:
        # default values: return relu_(x) = max(x, 0)
        return relu_(
            inputs, dc_decomp=dc_decomp, perturbation_domain=perturbation_domain, mode=mode, slope=slope, **kwargs
        )

    raise NotImplementedError()


def linear_hull_s_shape(
    inputs: List[tf.Tensor],
    func: Callable[[TensorLike], tf.Tensor] = K.sigmoid,
    f_prime: Callable[[TensorLike], tf.Tensor] = sigmoid_prime,
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    slope: Union[str, Slope] = Slope.V_SLOPE,
) -> List[tf.Tensor]:
    """Computing the linear hull of s-shape functions

    Args:
        inputs: list of input tensors
        func: the function (sigmoid, tanh, softsign...)
        f_prime: the derivative of the function (sigmoid_prime...)
        dc_decomp: boolean that indicates
        perturbation_domain: type of convex input domain (None or dict)
        mode: type of Forward propagation (ibp, affine, or hybrid)
        slope:
    whether we return a difference of convex decomposition of our layer

    Returns:
        the updated list of tensors
    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()

    mode = ForwardMode(mode)
    affine = get_affine(mode)
    ibp = get_ibp(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)
    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(
        inputs, compute_ibp_from_affine=False
    )
    dtype = x.dtype
    empty_tensor = inputs_outputs_spec.get_empty_tensor(dtype=dtype)

    if ibp:
        u_c_out = func(u_c)
        l_c_out = func(l_c)
    else:
        u_c_out, l_c_out = empty_tensor, empty_tensor

    if affine:
        w_u_0, b_u_0, w_l_0, b_l_0 = get_linear_hull_s_shape(
            inputs, func=func, f_prime=f_prime, perturbation_domain=perturbation_domain, mode=mode
        )
        if len(w_u.shape) == len(b_u.shape):
            # it happens with the convert function to spare memory footprint
            n_dim = np.prod(w_u.shape[1:])
            M = np.reshape(
                np.diag([K.cast(1, dtype=w_u_0.dtype)] * n_dim), [1, n_dim] + list(w_u.shape[1:])
            )  # usage de numpy pb pour les types
            w_u_out = M * K.concatenate([K.expand_dims(w_u_0, 1)] * n_dim, 1)
            w_l_out = M * K.concatenate([K.expand_dims(w_l_0, 1)] * n_dim, 1)
            b_u_out = b_u_0
            b_l_out = b_l_0
        else:
            w_u_out = K.expand_dims(w_u_0, 1) * w_u  # pour l'instant
            b_u_out = b_u_0 + w_u_0 * b_u
            w_l_out = K.expand_dims(w_l_0, 1) * w_l
            b_l_out = b_l_0 + w_l_0 * b_l
    else:
        w_u_out, b_u_out, w_l_out, b_l_out = empty_tensor, empty_tensor, empty_tensor, empty_tensor

    if dc_decomp:
        raise NotImplementedError()
    else:
        h_out, g_out = empty_tensor, empty_tensor

    return inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
        [x, u_c_out, w_u_out, b_u_out, l_c_out, w_l_out, b_l_out, h_out, g_out]
    )


def sigmoid(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """LiRPA for Sigmoid activation function .
    `1 / (1 + exp(-x))`.

    Args:
        inputs: list of input tensors
        dc_decomp: boolean that indicates
        perturbation_domain: type of convex input domain (None or dict)
        mode: type of Forward propagation (ibp, affine, or hybrid)
        **kwargs: see Keras official documentation
    whether we return a difference of convex decomposition of our layer

    Returns:
        the updated list of tensors
    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    func = K.sigmoid
    f_prime = sigmoid_prime
    return linear_hull_s_shape(
        inputs, func, f_prime, dc_decomp=dc_decomp, perturbation_domain=perturbation_domain, mode=mode, slope=slope
    )


def tanh(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """LiRPA for Hyperbolic activation function.
    `tanh(x)=2*sigmoid(2*x)+1`

    Args:
        inputs: list of input tensors
        dc_decomp: boolean that indicates
        perturbation_domain: type of convex input domain (None or dict)
        mode: type of Forward propagation (ibp, affine, or hybrid)
        slope:
        **kwargs: see Keras official documentation
    whether we return a difference of convex decomposition of our layer

    Returns:
        the updated list of tensors
    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    func = K.tanh
    f_prime = tanh_prime
    return linear_hull_s_shape(
        inputs, func, f_prime, dc_decomp=dc_decomp, perturbation_domain=perturbation_domain, mode=mode, slope=slope
    )


def hard_sigmoid(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """LiRPA for Hard sigmoid activation function.
       Faster to compute than sigmoid activation.

    Args:
        inputs: list of input tensors
        dc_decomp: boolean that indicates
        perturbation_domain: type of convex input domain (None or dict)
        mode: type of Forward propagation (ibp, affine, or hybrid)
        slope:
        **kwargs: see Keras official documentation
    whether we return a difference of convex decomposition of our layer

    Returns:
        the updated list of tensors
    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    if dc_decomp:
        raise NotImplementedError()
    mode = ForwardMode(mode)
    raise NotImplementedError()


def elu(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """LiRPA for Exponential linear unit.

    Args:
        inputs: list of input tensors
        dc_decomp: boolean that indicates
        perturbation_domain: type of convex input domain (None or dict)
        mode: type of Forward propagation (ibp, affine, or hybrid)
        slope:
        **kwargs: see Keras official documentation
    whether we return a difference of convex decomposition of our layer

    Returns:
        the updated list of tensors
    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    if dc_decomp:
        raise NotImplementedError()
    mode = ForwardMode(mode)
    raise NotImplementedError()


def selu(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """LiRPA for Scaled Exponential Linear Unit (SELU).

    SELU is equal to: `scale * elu(x, alpha)`, where alpha and scale
    are predefined constants. The values of `alpha` and `scale` are
    chosen so that the mean and variance of the inputs are preserved
    between two consecutive layers as long as the weights are initialized
    correctly (see `lecun_normal` initialization) and the number of inputs
    is "large enough" (see references for more information).

    Args:
        inputs: list of input tensors
        dc_decomp: boolean that indicates
        perturbation_domain: type of convex input domain (None or dict)
        mode: type of Forward propagation (ibp, affine, or hybrid)
        slope:
        **kwargs: see Keras official documentation
    whether we return a difference of convex decomposition of our layer

    Returns:
        the updated list of tensors
    """
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    if dc_decomp:
        raise NotImplementedError()
    mode = ForwardMode(mode)
    raise NotImplementedError()


def linear(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """LiRPA foe Linear (i.e. identity) activation function.

    Args:
        inputs: list of input tensors
        dc_decomp: boolean that indicates
        perturbation_domain: type of convex input domain (None or dict)
        mode: type of Forward propagation (ibp, affine, or hybrid)
        slope:
        **kwargs: see Keras official documentation
    whether we return a difference of convex decomposition of our layer

    Returns:
        the updated list of tensors
    """
    return inputs


def exponential(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """LiRPA for Exponential activation function.

    Args:
        inputs: list of input tensors
        dc_decomp: boolean that indicates
        perturbation_domain: type of convex input domain (None or dict)
        mode: type of Forward propagation (ibp, affine, or hybrid)
        slope:
        **kwargs: see Keras official documentation
    whether we return a difference of convex decomposition of our layer

    Returns:
        the updated list of tensors
    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    return exp(inputs, dc_decomp=dc_decomp, perturbation_domain=perturbation_domain, mode=mode, slope=slope, **kwargs)


def softplus(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """LiRPA for Softplus activation function `log(exp(x) + 1)`.

    Args:
        inputs: list of input tensors
        dc_decomp: boolean that indicates
        perturbation_domain: type of convex input domain (None or dict)
        mode: type of Forward propagation (ibp, affine, or hybrid)
        slope:
        **kwargs: see Keras official documentation
    whether we return a difference of convex decomposition of our layer

    Returns:
        the updated list of tensors
    """
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    if dc_decomp:
        raise NotImplementedError()

    return softplus_(inputs, dc_decomp=dc_decomp, perturbation_domain=perturbation_domain, mode=mode, slope=slope)


def softsign(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """LiRPA for Softsign activation function `x / (abs(x) + 1)`.

    Args:
        inputs: list of input tensors
        dc_decomp: boolean that indicates
        perturbation_domain: type of convex input domain (None or dict)
        mode: type of Forward propagation (ibp, affine, or hybrid)
        slope:
        **kwargs: see Keras official documentation
    whether we return a difference of convex decomposition of our layer

    Returns:
        the updated list of tensors
    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    func = K.softsign
    f_prime = softsign_prime
    return linear_hull_s_shape(
        inputs, func, f_prime, dc_decomp=dc_decomp, perturbation_domain=perturbation_domain, mode=mode, slope=slope
    )


def softmax(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    axis: int = -1,
    clip: bool = True,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """LiRPA for Softmax activation function.

    Args:
        inputs: list of input tensors
        dc_decomp: boolean that indicates
        perturbation_domain: type of convex input domain (None or dict)
        mode: type of Forward propagation (ibp, affine, or hybrid)
        slope:
        **kwargs: see Keras official documentation
    whether we return a difference of convex decomposition of our layer

    Returns:
        the updated list of tensors
    """
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    if dc_decomp:
        raise NotImplementedError()
    mode = ForwardMode(mode)
    ibp = get_ibp(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)

    outputs_exp = exponential(
        minus(inputs, mode=mode, perturbation_domain=perturbation_domain),
        dc_decomp=dc_decomp,
        perturbation_domain=perturbation_domain,
        mode=mode,
        slope=slope,
    )
    outputs = expand_dims(
        frac_pos(
            sum(outputs_exp, axis=axis, dc_decomp=dc_decomp, perturbation_domain=perturbation_domain, mode=mode),
            dc_decomp=dc_decomp,
            perturbation_domain=perturbation_domain,
            mode=mode,
        ),
        mode=mode,
        axis=axis,
        perturbation_domain=perturbation_domain,
    )
    outputs = multiply(outputs_exp, outputs, mode=mode, perturbation_domain=perturbation_domain)
    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(
        outputs, compute_ibp_from_affine=False
    )
    if ibp:
        u_c = K.minimum(u_c, 1.0)
        l_c = K.maximum(l_c, 0.0)

    return inputs_outputs_spec.extract_outputsformode_from_fulloutputs([x, u_c, w_u, b_u, l_c, w_l, b_l, h, g])


def group_sort_2(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    data_format: str = "channels_last",
    slope: Union[str, Slope] = Slope.V_SLOPE,
    **kwargs: Any,
) -> List[tf.Tensor]:
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    mode = ForwardMode(mode)
    raise NotImplementedError()


def deserialize(name: str) -> Callable[..., List[tf.Tensor]]:
    """Get the activation from name.

    Args:
        name: name of the method.
    among the implemented Keras activation function.

    Returns:
        the activation function
    """
    name = name.lower()

    if name == SOFTMAX:
        return softmax
    elif name == ELU:
        return elu
    elif name == SELU:
        return selu
    elif name == SOFTPLUS:
        return softplus
    elif name == SOFTSIGN:
        return softsign
    elif name == SIGMOID:
        return sigmoid
    elif name == TANH:
        return tanh
    elif name == RELU:
        return relu
    elif name == EXPONENTIAL:
        return exponential
    elif name == LINEAR:
        return linear
    elif name == GROUP_SORT_2:
        return group_sort_2
    else:
        raise ValueError(f"Could not interpret activation function identifier: {name}")


def get(identifier: Any) -> Callable[..., List[tf.Tensor]]:
    """Get the `identifier` activation function.

    Args:
        identifier: None or str, name of the function.

    Returns:
        The activation function, `linear` if `identifier` is None.

    """
    if identifier is None:
        return linear
    elif isinstance(identifier, str):
        return deserialize(identifier)
    elif callable(identifier):
        if isinstance(identifier, DecomonLayer):
            warnings.warn(
                "Do not pass a layer instance (such as {identifier}) as the "
                "activation argument of another layer. Instead, advanced "
                "activation layers should be used just like any other "
                "layer in a model.".format(identifier=identifier.__class__.__name__)
            )
        return identifier
    else:
        raise ValueError("Could not interpret " "activation function identifier:", identifier)



from copy import deepcopy
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.layers import Concatenate, Lambda, Layer
from tensorflow.keras.models import Model
from tensorflow.python.keras.utils.generic_utils import to_list

from decomon.backward_layers.backward_merge import BackwardMerge
from decomon.backward_layers.convert import to_backward
from decomon.backward_layers.core import BackwardLayer
from decomon.core import (
    BoxDomain,
    ForwardMode,
    InputsOutputsSpec,
    PerturbationDomain,
    Slope,
    get_affine,
    get_mode,
)
from decomon.layers.utils import softmax_to_linear as softmax_2_linear
from decomon.models.crown import Convert2BackwardMode, Fuse, MergeWithPrevious
from decomon.models.forward_cloning import OutputMapDict
from decomon.models.utils import Convert2Mode, get_depth_dict

try:
    from keras.src.engine.node import Node  # new path starting from keras 2.13
except ImportError:
    from keras.engine.node import Node  # old path until keras 2.12


def get_disconnected_input(
    mode: Union[str, ForwardMode],
    perturbation_domain: PerturbationDomain,
    dtype: Union[str, tf.DType] = K.floatx(),
) -> Layer:
    mode = ForwardMode(mode)
    dc_decomp = False
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)
    affine = get_affine(mode)

    def disco_priv(inputs: List[tf.Tensor]) -> List[tf.Tensor]:
        x, u_c, w_f_u, b_f_u, l_c, w_f_l, b_f_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(inputs)
        dtype = x.dtype
        empty_tensor = inputs_outputs_spec.get_empty_tensor(dtype=dtype)

        if affine:
            x = K.concatenate([K.expand_dims(l_c, 1), K.expand_dims(u_c, 1)], 1)
            w_u = tf.linalg.diag(K.cast(0.0, x.dtype) * u_c + K.cast(1.0, x.dtype))
            b_u = K.cast(0.0, x.dtype) * u_c
        else:
            w_u, b_u = empty_tensor, empty_tensor

        return inputs_outputs_spec.extract_outputsformode_from_fulloutputs([x, u_c, w_u, b_u, l_c, w_u, b_u])

    return Lambda(disco_priv, dtype=dtype)


def retrieve_layer(
    node: Node,
    layer_fn: Callable[[Layer], BackwardLayer],
    backward_map: Dict[int, BackwardLayer],
    joint: bool = True,
) -> BackwardLayer:
    if id(node) in backward_map:
        backward_layer = backward_map[id(node)]
    else:
        backward_layer = layer_fn(node.outbound_layer)
        if joint:
            backward_map[id(node)] = backward_layer
    return backward_layer


def crown_(
    node: Node,
    ibp: bool,
    affine: bool,
    perturbation_domain: PerturbationDomain,
    input_map: Dict[int, List[tf.Tensor]],
    layer_fn: Callable[[Layer], BackwardLayer],
    backward_bounds: List[tf.Tensor],
    backward_map: Optional[Dict[int, BackwardLayer]] = None,
    joint: bool = True,
    fuse: bool = True,
    output_map: Optional[Dict[int, List[tf.Tensor]]] = None,
    merge_layers: Optional[Layer] = None,
    fuse_layer: Optional[Layer] = None,
    **kwargs: Any,
) -> Tuple[List[tf.Tensor], Optional[Layer]]:
    """


    :param node:
    :param ibp:
    :param affine:
    :param input_map:
    :param layer_fn:
    :param backward_bounds:
    :param backward_map:
    :param joint:
    :param fuse:
    :return: list of 4 tensors affine upper and lower bounds
    """
    if backward_map is None:
        backward_map = {}

    if output_map is None:
        output_map = {}

    inputs = input_map[id(node)]

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()

    if isinstance(node.outbound_layer, Model):
        inputs_tensors = get_disconnected_input(get_mode(ibp, affine), perturbation_domain, dtype=inputs[0].dtype)(
            inputs
        )
        _, backward_bounds, _, _ = crown_model(
            model=node.outbound_layer,
            input_tensors=inputs_tensors,
            backward_bounds=backward_bounds,
            ibp=ibp,
            affine=affine,
            perturbation_domain=None,
            finetune=False,
            joint=joint,
            fuse=False,
            **kwargs,
        )

    else:
        backward_layer = retrieve_layer(node=node, layer_fn=layer_fn, backward_map=backward_map, joint=joint)

        if id(node) not in output_map:
            backward_bounds_new = backward_layer(inputs)
            output_map[id(node)] = backward_bounds_new
        else:
            backward_bounds_new = output_map[id(node)]

            # import pdb; pdb.set_trace()
        if len(backward_bounds):
            if merge_layers is None:
                merge_layers = MergeWithPrevious(backward_bounds_new[0].shape, backward_bounds[0].shape)
            backward_bounds = merge_layers(backward_bounds_new + backward_bounds)
        else:
            backward_bounds = backward_bounds_new

    parents = node.parent_nodes

    if len(parents):
        if len(parents) > 1:
            if isinstance(backward_layer, BackwardMerge):
                raise NotImplementedError()
                crown_bound_list = []
                for backward_bound, parent in zip(backward_bounds, parents):
                    crown_bound_i, _ = crown_(
                        node=parent,
                        ibp=ibp,
                        affine=affine,
                        perturbation_domain=perturbation_domain,
                        input_map=input_map,
                        layer_fn=layer_fn,
                        backward_bounds=backward_bound,
                        backward_map=backward_map,
                        joint=joint,
                        fuse=fuse,
                    )

                    crown_bound_list.append(crown_bound_i)

                # avg_layer = Average(dtype=node.outbound_layer.dtype)
                # crown_bound = [avg_layer([e[i] for e in crown_bound_list]) for i in range(4)]
                crown_bound = crown_bound_list[0]

            else:
                raise NotImplementedError()
        else:
            crown_bound, fuse_layer_new = crown_(
                node=parents[0],
                ibp=ibp,
                affine=affine,
                perturbation_domain=perturbation_domain,
                input_map=input_map,
                layer_fn=layer_fn,
                backward_bounds=backward_bounds,
                backward_map=backward_map,
                joint=joint,
                fuse=fuse,
                output_map=output_map,
                merge_layers=None,  # AKA merge_layers
                fuse_layer=fuse_layer,
            )
            if fuse_layer is None:
                fuse_layer = fuse_layer_new
        return crown_bound, fuse_layer
    else:
        # do something
        if fuse:
            if fuse_layer is None:
                fuse_layer = Fuse(get_mode(ibp=ibp, affine=affine))
            result = fuse_layer(inputs + backward_bounds)

            return result, fuse_layer

        else:
            return backward_bounds, fuse_layer


def get_input_nodes(
    model: Model,
    dico_nodes: Dict[int, List[Node]],
    ibp: bool,
    affine: bool,
    input_tensors: List[tf.Tensor],
    output_map: OutputMapDict,
    layer_fn: Callable[[Layer], BackwardLayer],
    joint: bool,
    set_mode_layer: Layer,
    perturbation_domain: Optional[PerturbationDomain] = None,
    **kwargs: Any,
) -> Tuple[Dict[int, List[tf.Tensor]], Dict[int, BackwardLayer], Dict[int, List[tf.Tensor]]]:
    keys = [e for e in dico_nodes.keys()]
    keys.sort(reverse=True)
    fuse_layer = None
    input_map: Dict[int, List[tf.Tensor]] = {}
    backward_map: Dict[int, BackwardLayer] = {}
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    crown_map: Dict[int, List[tf.Tensor]] = {}
    for depth in keys:
        nodes = dico_nodes[depth]
        for node in nodes:
            layer = node.outbound_layer

            parents = node.parent_nodes
            if not len(parents):
                # if 'debug' in kwargs.keys():
                #    import pdb; pdb.set_trace()
                input_map[id(node)] = input_tensors
            else:
                output: List[tf.Tensor] = []
                for parent in parents:
                    # do something
                    if id(parent) in output_map.keys():
                        output += output_map[id(parent)]
                    else:
                        output_crown, fuse_layer_tmp = crown_(
                            node=parent,
                            ibp=ibp,
                            affine=affine,
                            input_map=input_map,
                            layer_fn=layer_fn,
                            backward_bounds=[],
                            backward_map=backward_map,
                            joint=joint,
                            fuse=True,
                            perturbation_domain=perturbation_domain,
                            output_map=crown_map,
                            merge_layers=None,  # AKA merge_layers
                            fuse_layer=fuse_layer,
                        )
                        if fuse_layer is None:
                            fuse_layer = fuse_layer_tmp

                        # convert output_crown in the right mode
                        if set_mode_layer is None:
                            set_mode_layer = Convert2BackwardMode(get_mode(ibp, affine), perturbation_domain)
                        output_crown = set_mode_layer(input_tensors + output_crown)
                        output += to_list(output_crown)
                        # crown_map[id(parent)]=output_crown_

                input_map[id(node)] = output
    return input_map, backward_map, crown_map


def crown_model(
    model: Model,
    input_tensors: List[tf.Tensor],
    back_bounds: Optional[List[tf.Tensor]] = None,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    ibp: bool = True,
    affine: bool = True,
    perturbation_domain: Optional[PerturbationDomain] = None,
    finetune: bool = False,
    forward_map: Optional[OutputMapDict] = None,
    softmax_to_linear: bool = True,
    joint: bool = True,
    layer_fn: Callable[..., BackwardLayer] = to_backward,
    fuse: bool = True,
    **kwargs: Any,
) -> Tuple[List[tf.Tensor], List[tf.Tensor], Dict[int, BackwardLayer], None]:
    if back_bounds is None:
        back_bounds = []
    if forward_map is None:
        forward_map = {}
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    if not isinstance(model, Model):
        raise ValueError()
    # import time
    # zero_time = time.process_time()
    has_softmax = False
    if softmax_to_linear:
        model, has_softmax = softmax_2_linear(model)  # do better because you modify the model eventually

    # layer_fn
    ##########
    has_iter = False
    if layer_fn is not None and len(layer_fn.__code__.co_varnames) == 1 and "layer" in layer_fn.__code__.co_varnames:
        has_iter = True

    if not has_iter:
        layer_fn_copy = deepcopy(layer_fn)

        def func(layer: Layer) -> Layer:
            return layer_fn_copy(
                layer,
                mode=get_mode(ibp, affine),
                finetune=finetune,
                perturbation_domain=perturbation_domain,
                slope=slope,
            )

        layer_fn = func

    if not callable(layer_fn):
        raise ValueError("Expected `layer_fn` argument to be a callable.")
    ###############

    if len(back_bounds) and len(to_list(model.output)) > 1:
        raise NotImplementedError()

    # sort nodes from input to output
    dico_nodes = get_depth_dict(model)
    keys = [e for e in dico_nodes.keys()]
    keys.sort(reverse=True)

    # generate input_map
    if not finetune:
        joint = True
    set_mode_layer = Convert2BackwardMode(get_mode(ibp, affine), perturbation_domain)

    input_map, backward_map, crown_map = get_input_nodes(
        model=model,
        dico_nodes=dico_nodes,
        ibp=ibp,
        affine=affine,
        input_tensors=input_tensors,
        output_map=forward_map,
        layer_fn=layer_fn,
        joint=joint,
        perturbation_domain=perturbation_domain,
        set_mode_layer=set_mode_layer,
        **kwargs,
    )
    # time_1 = time.process_time()
    # print('step1', time_1-zero_time)
    # retrieve output nodes
    output = []
    output_nodes = dico_nodes[0]
    # the ordering may change
    output_names = [tensor._keras_history.layer.name for tensor in to_list(model.output)]
    fuse_layer = None
    for output_name in output_names:
        for node in output_nodes:
            if node.outbound_layer.name == output_name:
                # compute with crown
                output_crown, fuse_layer = crown_(
                    node=node,
                    ibp=ibp,
                    affine=affine,
                    input_map=input_map,
                    layer_fn=layer_fn,
                    backward_bounds=back_bounds,
                    backward_map=backward_map,
                    joint=joint,
                    fuse=fuse,
                    perturbation_domain=perturbation_domain,
                    output_map=crown_map,
                    fuse_layer=fuse_layer,
                )
                # time_2 = time.process_time()
                # print('step2', time_2-time_1)
                if fuse:
                    # import pdb; pdb.set_trace()
                    output += to_list(set_mode_layer(input_tensors + output_crown))
                else:
                    output = output_crown

    return input_tensors, output, backward_map, None


def convert_backward(
    model: Model,
    input_tensors: List[tf.Tensor],
    back_bounds: Optional[List[tf.Tensor]] = None,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    ibp: bool = True,
    affine: bool = True,
    perturbation_domain: Optional[PerturbationDomain] = None,
    finetune: bool = False,
    forward_map: Optional[OutputMapDict] = None,
    softmax_to_linear: bool = True,
    joint: bool = True,
    layer_fn: Callable[..., BackwardLayer] = to_backward,
    final_ibp: bool = True,
    final_affine: bool = False,
    **kwargs: Any,
) -> Tuple[List[tf.Tensor], List[tf.Tensor], Dict[int, BackwardLayer], None]:
    if back_bounds is None:
        back_bounds = []
    if forward_map is None:
        forward_map = {}
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    if len(back_bounds):
        if len(back_bounds) == 1:
            C = back_bounds[0]
            bias = K.cast(0.0, model.layers[0].dtype) * C[:, 0]
            back_bounds = [C, bias] * 2
    result = crown_model(
        model=model,
        input_tensors=input_tensors,
        back_bounds=back_bounds,
        slope=slope,
        ibp=ibp,
        affine=affine,
        perturbation_domain=perturbation_domain,
        finetune=finetune,
        forward_map=forward_map,
        softmax_to_linear=softmax_to_linear,
        joint=joint,
        layer_fn=layer_fn,
        fuse=True,
        **kwargs,
    )

    input_tensors, output, backward_map, toto = result
    mode_from = get_mode(ibp, affine)
    mode_to = get_mode(final_ibp, final_affine)
    output = Convert2Mode(
        mode_from=mode_from,
        mode_to=mode_to,
        perturbation_domain=perturbation_domain,
    )(output)
    if mode_to != mode_from and mode_from == ForwardMode.IBP:
        f_input = Lambda(lambda z: Concatenate(1)([z[0][:, None], z[1][:, None]]))
        output[0] = f_input([input_tensors[1], input_tensors[0]])
    return input_tensors, output, backward_map, toto



"""Module for DecomonSequential.

It inherits from keras Sequential class.

"""
import inspect
from copy import deepcopy
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import tensorflow as tf
from tensorflow.keras.layers import Layer
from tensorflow.keras.models import Model
from tensorflow.python.keras.utils.generic_utils import to_list

from decomon.core import BoxDomain, PerturbationDomain, Slope
from decomon.layers.convert import to_decomon
from decomon.layers.core import DecomonLayer
from decomon.layers.utils import softmax_to_linear as softmax_2_linear
from decomon.models.utils import (
    get_depth_dict,
    get_inner_layers,
    get_input_dim,
    prepare_inputs_for_layer,
    wrap_outputs_from_layer_in_list,
)

OutputMapKey = Union[str, int]
OutputMapVal = Union[List[tf.Tensor], "OutputMapDict"]
OutputMapDict = Dict[OutputMapKey, OutputMapVal]

LayerMapVal = Union[List[DecomonLayer], "LayerMapDict"]
LayerMapDict = Dict[int, LayerMapVal]


def include_dim_layer_fn(
    layer_fn: Callable[..., Layer],
    input_dim: int,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    ibp: bool = True,
    affine: bool = True,
    finetune: bool = False,
    shared: bool = True,
) -> Callable[[Layer], List[Layer]]:
    """include external parameters inside the translation of a layer to its decomon counterpart

    Args:
        layer_fn
        input_dim
        dc_decomp
        perturbation_domain
        finetune

    Returns:

    """
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()

    if not callable(layer_fn):
        raise ValueError("Expected `layer_fn` argument to be a callable.")

    layer_fn_copy = deepcopy(layer_fn)

    if "input_dim" in inspect.signature(layer_fn).parameters:

        def func(layer: Layer) -> List[Layer]:
            return [
                layer_fn_copy(
                    layer,
                    input_dim=input_dim,
                    slope=slope,
                    perturbation_domain=perturbation_domain,
                    dc_decomp=dc_decomp,
                    ibp=ibp,
                    affine=affine,
                    finetune=finetune,
                    shared=shared,
                )
            ]

    else:

        def func(layer: Layer) -> List[Layer]:
            return [layer_fn_copy(layer)]

    return func


def convert_forward(
    model: Model,
    input_tensors: List[tf.Tensor],
    layer_fn: Callable[..., Layer] = to_decomon,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    input_dim: int = -1,
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    ibp: bool = True,
    affine: bool = True,
    finetune: bool = False,
    shared: bool = True,
    softmax_to_linear: bool = True,
    joint: bool = True,
    **kwargs: Any,
) -> Tuple[List[tf.Tensor], List[tf.Tensor], LayerMapDict, OutputMapDict]:
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()

    if not isinstance(model, Model):
        raise ValueError("Expected `model` argument " "to be a `Model` instance, got ", model)

    if input_dim == -1:
        input_dim = get_input_dim(model)

    layer_fn_to_list = include_dim_layer_fn(
        layer_fn,
        input_dim=input_dim,
        slope=slope,
        perturbation_domain=perturbation_domain,
        ibp=ibp,
        affine=affine,
        finetune=finetune,
        shared=shared,
    )  # return a list of Decomon layers

    f_output = convert_forward_functional_model(
        model=model,
        input_tensors=input_tensors,
        layer_fn=layer_fn_to_list,
        softmax_to_linear=softmax_to_linear,
        joint=joint,
    )

    return f_output


def convert_forward_functional_model(
    model: Model,
    layer_fn: Callable[[Layer], List[Layer]],
    input_tensors: List[tf.Tensor],
    softmax_to_linear: bool = True,
    count: int = 0,
    joint: bool = True,
    output_map: Optional[OutputMapDict] = None,
    layer_map: Optional[LayerMapDict] = None,
) -> Tuple[List[tf.Tensor], List[tf.Tensor], LayerMapDict, OutputMapDict]:
    if softmax_to_linear:
        model, has_softmax = softmax_2_linear(model)

    # create input tensors

    # sort nodes from input to output
    dico_nodes = get_depth_dict(model)
    keys = [e for e in dico_nodes.keys()]
    keys.sort(reverse=True)

    if output_map is None:
        output_map = {}
    if layer_map is None:
        layer_map = {}
    output: List[tf.Tensor] = input_tensors
    for depth in keys:
        nodes = dico_nodes[depth]
        for node in nodes:
            layer = node.outbound_layer
            parents = node.parent_nodes
            if id(node) in output_map.keys() and joint:
                continue
            if len(parents):
                output = []
                for parent in parents:
                    output += output_map[id(parent)]

            if isinstance(layer, Model):
                _, output, layer_map_submodel, output_map_submodel = convert_forward_functional_model(
                    model=layer,
                    input_tensors=output,
                    layer_fn=layer_fn,
                    softmax_to_linear=softmax_to_linear,
                    count=count,
                )
                count = count + get_inner_layers(layer)
                layer_map[id(node)] = layer_map_submodel
                output_map[id(node)] = output_map_submodel
            else:
                converted_layers = layer_fn(layer)
                for converted_layer in converted_layers:
                    converted_layer._name = f"{converted_layer.name}_{count}"
                    count += 1
                    output = converted_layer(prepare_inputs_for_layer(output))
                    if len(converted_layers) > 1:
                        output_map[f"{id(node)}_{converted_layer.name}"] = wrap_outputs_from_layer_in_list(output)
                layer_map[id(node)] = converted_layers
            output_map[id(node)] = wrap_outputs_from_layer_in_list(output)

    output = []
    output_nodes = dico_nodes[0]
    # the ordering may change
    output_names = [tensor._keras_history.layer.name for tensor in to_list(model.output)]
    for output_name in output_names:
        for node in output_nodes:
            if node.outbound_layer.name == output_name:
                output += output_map[id(node)]

    return input_tensors, output, layer_map, output_map



import numpy as np
import pytest
import tensorflow as tf
from numpy.testing import assert_almost_equal
from tensorflow.keras.layers import BatchNormalization, Conv2D, Dense, Flatten

from decomon.core import ForwardMode
from decomon.layers.decomon_layers import (
    DecomonBatchNormalization,
    DecomonConv2D,
    DecomonDense,
    DecomonFlatten,
)


def test_decomondense_reset_layer(helpers, use_bias):
    dc_decomp = False
    input_dim = 1
    units = 3
    mode = ForwardMode.HYBRID
    layer = Dense(units=units, use_bias=use_bias)
    layer(tf.zeros((2, input_dim)))
    decomon_layer = DecomonDense(units=units, use_bias=use_bias, mode=mode, dc_decomp=dc_decomp)
    inputs = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    decomon_layer(inputs_for_mode)

    kernel = layer.kernel
    layer.kernel.assign(2 * np.ones_like(kernel))
    decomon_layer.kernel.assign(np.zeros_like(kernel))
    if use_bias:
        bias = layer.bias
        layer.bias.assign(np.ones_like(bias))
        decomon_layer.bias.assign(np.zeros_like(bias))

    decomon_layer.reset_layer(layer)
    assert decomon_layer.kernel is not layer.kernel
    assert_almost_equal(decomon_layer.kernel.numpy(), layer.kernel.numpy())
    if use_bias:
        assert len(layer.weights) == 2
        assert decomon_layer.bias is not layer.bias
        assert_almost_equal(decomon_layer.bias.numpy(), layer.bias.numpy())
    else:
        assert len(layer.weights) == 1


def test_decomondense_reset_layer_decomon_with_new_weights(helpers):
    dc_decomp = False
    input_dim = 1
    units = 3
    mode = ForwardMode.HYBRID
    layer = Dense(units=units)
    layer(tf.zeros((2, input_dim)))
    decomon_layer = DecomonDense(units=units, mode=mode, dc_decomp=dc_decomp)
    inputs = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    decomon_layer(inputs_for_mode)
    # Add new variables
    decomon_layer.add_weight(shape=(input_dim, units), initializer="ones", name="alpha", trainable=True)
    decomon_layer.add_weight(shape=(input_dim, units), initializer="ones", name="beta", trainable=False)
    assert len(decomon_layer.weights) == 4
    assert len(decomon_layer.trainable_weights) == 3

    kernel, bias = layer.get_weights()
    layer.set_weights([2 * np.ones_like(kernel), np.ones_like(bias)])
    decomon_layer.kernel.assign(np.zeros_like(kernel))
    decomon_layer.bias.assign(np.zeros_like(bias))

    decomon_layer.reset_layer(layer)
    assert decomon_layer.kernel is not layer.kernel
    assert decomon_layer.bias is not layer.bias
    assert_almost_equal(decomon_layer.kernel.numpy(), layer.kernel.numpy())
    assert_almost_equal(decomon_layer.bias.numpy(), layer.bias.numpy())


def test_decomondense_reset_layer_keras_with_new_weights(helpers):
    dc_decomp = False
    input_dim = 1
    units = 3
    mode = ForwardMode.HYBRID
    layer = Dense(units=units)
    layer(tf.zeros((2, input_dim)))
    decomon_layer = DecomonDense(units=units, mode=mode, dc_decomp=dc_decomp)
    inputs = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    decomon_layer(inputs_for_mode)
    # Add new variables
    layer.add_weight(shape=(input_dim, units), initializer="ones", name="alpha", trainable=False)
    assert len(layer.weights) == 3

    kernel = layer.kernel
    layer.kernel.assign(2 * np.ones_like(kernel))
    decomon_layer.kernel.assign(np.zeros_like(kernel))
    bias = layer.bias
    layer.bias.assign(np.ones_like(bias))
    decomon_layer.bias.assign(np.zeros_like(bias))

    decomon_layer.reset_layer(layer)
    assert decomon_layer.kernel is not layer.kernel
    assert decomon_layer.bias is not layer.bias
    assert_almost_equal(decomon_layer.kernel.numpy(), layer.kernel.numpy())
    assert_almost_equal(decomon_layer.bias.numpy(), layer.bias.numpy())


def test_decomondense_reset_layer_ko_keraslayer_not_nuilt():
    dc_decomp = False
    input_dim = 1
    units = 3
    mode = ForwardMode.HYBRID
    layer = Dense(units=units)
    decomon_layer = DecomonDense(units=units, mode=mode, dc_decomp=dc_decomp)
    with pytest.raises(ValueError):
        decomon_layer.reset_layer(layer)


def test_decomondense_reset_layer_ko_decomonlayer_not_nuilt():
    dc_decomp = False
    input_dim = 1
    units = 3
    mode = ForwardMode.HYBRID
    layer = Dense(units=units)
    layer(tf.zeros((2, input_dim)))
    decomon_layer = DecomonDense(units=units, mode=mode, dc_decomp=dc_decomp)
    with pytest.raises(ValueError):
        decomon_layer.reset_layer(layer)


def test_decomonconv2d_reset_layer(helpers, use_bias):
    dc_decomp = False
    odd = 0
    data_format = "channels_last"
    filters = 3
    kernel_size = (3, 3)
    mode = ForwardMode.HYBRID

    inputs = helpers.get_tensor_decomposition_images_box(data_format, odd, dc_decomp=False)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    input_ref = helpers.get_input_ref_from_full_inputs(inputs=inputs)

    layer = Conv2D(filters=filters, kernel_size=kernel_size, use_bias=use_bias)
    layer(input_ref)
    decomon_layer = DecomonConv2D(
        filters=filters, kernel_size=kernel_size, use_bias=use_bias, mode=mode, dc_decomp=dc_decomp
    )
    decomon_layer(inputs_for_mode)

    kernel = layer.kernel
    layer.kernel.assign(2 * np.ones_like(kernel))
    decomon_layer.kernel.assign(np.zeros_like(kernel))
    if use_bias:
        bias = layer.bias
        layer.bias.assign(np.ones_like(bias))
        decomon_layer.bias.assign(np.zeros_like(bias))

    decomon_layer.reset_layer(layer)
    assert decomon_layer.kernel is not layer.kernel
    assert_almost_equal(decomon_layer.kernel.numpy(), layer.kernel.numpy())
    if use_bias:
        assert decomon_layer.bias is not layer.bias
        assert_almost_equal(decomon_layer.bias.numpy(), layer.bias.numpy())


@pytest.mark.parametrize(
    "center, scale",
    [
        (True, True),
        (False, False),
    ],
)
def test_decomonbacthnormalization_reset_layer(helpers, center, scale):
    dc_decomp = False
    odd = 0
    mode = ForwardMode.HYBRID
    inputs = helpers.get_tensor_decomposition_multid_box(odd=odd, dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    input_ref = helpers.get_input_ref_from_full_inputs(inputs=inputs)

    layer = BatchNormalization(center=center, scale=scale)
    layer(input_ref)
    decomon_layer = DecomonBatchNormalization(center=center, scale=scale, mode=mode, dc_decomp=dc_decomp)
    decomon_layer(inputs_for_mode)

    moving_mean = layer.moving_mean
    layer.moving_mean.assign(np.ones_like(moving_mean))
    decomon_layer.moving_mean.assign(np.zeros_like(moving_mean))
    moving_variance = layer.moving_variance
    layer.moving_variance.assign(np.ones_like(moving_variance))
    decomon_layer.moving_variance.assign(np.zeros_like(moving_variance))
    if center:
        beta = layer.beta
        layer.beta.assign(np.ones_like(beta))
        decomon_layer.beta.assign(np.zeros_like(beta))
    if scale:
        gamma = layer.gamma
        layer.gamma.assign(np.ones_like(gamma))
        decomon_layer.gamma.assign(np.zeros_like(gamma))

    decomon_layer.reset_layer(layer)

    keras_weights = layer.get_weights()
    decomon_weights = layer.get_weights()
    for i in range(len(decomon_weights)):
        assert_almost_equal(decomon_weights[i], keras_weights[i])


def test_decomonflatten_reset_layer(helpers):
    dc_decomp = False
    input_dim = 1
    mode = ForwardMode.HYBRID
    layer = Flatten()
    layer(tf.zeros((2, input_dim)))
    decomon_layer = DecomonFlatten(mode=mode, dc_decomp=dc_decomp)
    inputs = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    decomon_layer(inputs_for_mode)

    decomon_layer.reset_layer(layer)



from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.backend import conv2d
from tensorflow.keras.layers import InputSpec, MaxPooling2D

from decomon.core import ForwardMode, PerturbationDomain
from decomon.layers.core import DecomonLayer
from decomon.layers.utils import max_


# step 1: compute the maximum
class DecomonMaxPooling2D(DecomonLayer, MaxPooling2D):
    """LiRPA implementation of MaxPooling2D layers.
    See Keras official documentation for further details on the MaxPooling2D operator

    """

    original_keras_layer_class = MaxPooling2D

    pool_size: Tuple[int, int]
    strides: Tuple[int, int]
    padding: str
    data_format: str

    def __init__(
        self,
        pool_size: Union[int, Tuple[int, int]] = (2, 2),
        strides: Optional[Union[int, Tuple[int, int]]] = None,
        padding: str = "valid",
        data_format: Optional[str] = None,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        finetune: bool = False,
        shared: bool = False,
        fast: bool = True,
        **kwargs: Any,
    ):
        super().__init__(
            pool_size=pool_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            mode=mode,
            finetune=finetune,
            shared=shared,
            fast=fast,
            **kwargs,
        )

        if self.mode == ForwardMode.IBP:
            self.input_spec = [
                InputSpec(ndim=4),  # u
                InputSpec(ndim=4),  # l
            ]
        elif self.mode == ForwardMode.AFFINE:
            self.input_spec = [
                InputSpec(min_ndim=2),  # x
                InputSpec(ndim=5),  # w_u
                InputSpec(ndim=4),  # b_u
                InputSpec(ndim=5),  # w_l
                InputSpec(ndim=4),
            ]  # b_l
        elif self.mode == ForwardMode.HYBRID:
            self.input_spec = [
                InputSpec(min_ndim=2),  # x
                InputSpec(ndim=4),  # u
                InputSpec(ndim=5),  # w_u
                InputSpec(ndim=4),  # b_u
                InputSpec(ndim=4),  # l
                InputSpec(ndim=5),  # w_l
                InputSpec(ndim=4),
            ]  # b_l
        else:
            raise ValueError(f"Unknown mode {self.mode}.")

        if self.dc_decomp:
            self.input_spec += [InputSpec(ndim=4), InputSpec(ndim=4)]

        # express maxpooling with convolutions
        self.filters = np.zeros((self.pool_size[0], self.pool_size[1], 1, np.prod(self.pool_size)), dtype=self.dtype)
        for i in range(self.pool_size[0]):
            for j in range(self.pool_size[1]):
                self.filters[i, j, 0, i * self.pool_size[0] + j] = 1

        def conv_(x: tf.Tensor) -> tf.Tensor:
            if self.data_format in [None, "channels_last"]:
                return K.cast(
                    K.expand_dims(
                        conv2d(
                            x,
                            self.filters,
                            strides=self.strides,
                            padding=self.padding,
                            data_format=self.data_format,
                        ),
                        -2,
                    ),
                    self.dtype,
                )
            else:
                return K.cast(
                    K.expand_dims(
                        conv2d(
                            x,
                            self.filters,
                            strides=self.strides,
                            padding=self.padding,
                            data_format=self.data_format,
                        ),
                        1,
                    ),
                    self.dtype,
                )

        self.internal_op = conv_

    def compute_output_shape(self, input_shape: List[tf.TensorShape]) -> List[tf.TensorShape]:
        """
        Args:
            input_shape

        Returns:

        """

        if self.grad_bounds:
            raise NotImplementedError()

        output_shape_keras = MaxPooling2D.compute_output_shape(self, input_shape[0])
        input_dim = input_shape[-1][1]
        if self.mode == ForwardMode.IBP:
            output_shape = [output_shape_keras] * 2
        elif self.mode in [ForwardMode.HYBRID, ForwardMode.AFFINE]:
            x_shape = input_shape[1]
            w_shape = tuple([output_shape_keras[0], input_dim] + list(output_shape_keras)[1:])
            if self.mode == ForwardMode.AFFINE:
                output_shape = [x_shape] + [w_shape, output_shape_keras] * 2
            else:
                output_shape = [x_shape] + [output_shape_keras, w_shape, output_shape_keras] * 2
        else:
            raise ValueError(f"Unknown mode {self.mode}")

        if self.dc_decomp:
            output_shape += [output_shape_keras] * 2

        return output_shape

    def _pooling_function_fast(
        self,
        inputs: List[tf.Tensor],
    ) -> List[tf.Tensor]:
        x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = self.inputs_outputs_spec.get_fullinputs_from_inputsformode(inputs)
        dtype = x.dtype
        empty_tensor = self.inputs_outputs_spec.get_empty_tensor(dtype=dtype)

        l_c_out = K.pool2d(l_c, self.pool_size, self.strides, self.padding, self.data_format, pool_mode="max")
        u_c_out = K.pool2d(u_c, self.pool_size, self.strides, self.padding, self.data_format, pool_mode="max")

        if self.affine:
            n_in = x.shape[-1]
            w_u_out = K.concatenate([0 * K.expand_dims(u_c_out, 1)] * n_in, 1)
            w_l_out = w_u_out
            b_u_out = u_c_out
            b_l_out = l_c_out
        else:
            w_u_out, b_u_out, w_l_out, b_l_out = empty_tensor, empty_tensor, empty_tensor, empty_tensor

        if self.dc_decomp:
            raise NotImplementedError()
        else:
            h_out, g_out = empty_tensor, empty_tensor

        return self.inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
            [x, u_c_out, w_u_out, b_u_out, l_c_out, w_l_out, b_l_out, h_out, g_out]
        )

    def _pooling_function_not_fast(
        self,
        inputs: List[tf.Tensor],
    ) -> List[tf.Tensor]:
        x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = self.inputs_outputs_spec.get_fullinputs_from_inputsformode(
            inputs, compute_ibp_from_affine=False
        )
        dtype = x.dtype
        empty_tensor = self.inputs_outputs_spec.get_empty_tensor(dtype=dtype)
        input_shape = self.inputs_outputs_spec.get_input_shape(inputs)
        n_split = input_shape[-1]

        if self.dc_decomp:
            h_out = K.concatenate(
                [self.internal_op(elem) for elem in tf.split(h, n_split, -1)],
                -2,
            )
            g_out = K.concatenate(
                [self.internal_op(elem) for elem in tf.split(g, n_split, -1)],
                -2,
            )
        else:
            h_out, g_out = empty_tensor, empty_tensor

        if self.ibp:
            u_c_out = K.concatenate([self.internal_op(elem) for elem in tf.split(u_c, n_split, -1)], -2)
            l_c_out = K.concatenate([self.internal_op(elem) for elem in tf.split(l_c, n_split, -1)], -2)
        else:
            u_c_out, l_c_out = empty_tensor, empty_tensor

        if self.affine:
            b_u_out = K.concatenate([self.internal_op(elem) for elem in tf.split(b_u, n_split, -1)], -2)
            b_l_out = K.concatenate([self.internal_op(elem) for elem in tf.split(b_l, n_split, -1)], -2)
            w_u_out = K.concatenate([self.internal_op(elem) for elem in tf.split(w_u, n_split, -1)], -2)
            w_l_out = K.concatenate([self.internal_op(elem) for elem in tf.split(w_l, n_split, -1)], -2)
        else:
            w_u_out, b_u_out, w_l_out, b_l_out = empty_tensor, empty_tensor, empty_tensor, empty_tensor

        outputs = self.inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
            [x, u_c_out, w_u_out, b_u_out, l_c_out, w_l_out, b_l_out, h_out, g_out]
        )

        return max_(
            outputs, axis=-1, dc_decomp=self.dc_decomp, mode=self.mode, perturbation_domain=self.perturbation_domain
        )

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        if self.fast:
            return self._pooling_function_fast(inputs)
        else:
            return self._pooling_function_not_fast(inputs)


# Aliases
DecomonMaxPool2d = DecomonMaxPooling2D



from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.backend import conv2d_transpose
from tensorflow.keras.layers import Flatten, Layer
from tensorflow.python.ops import array_ops

from decomon.backward_layers.activations import get
from decomon.backward_layers.core import BackwardLayer
from decomon.backward_layers.utils import get_identity_lirpa
from decomon.backward_layers.utils_conv import get_toeplitz
from decomon.core import (
    ForwardMode,
    GridDomain,
    Option,
    PerturbationDomain,
    Slope,
    get_affine,
    get_ibp,
)
from decomon.layers.convert import to_decomon
from decomon.layers.core import DecomonLayer
from decomon.layers.decomon_layers import DecomonBatchNormalization
from decomon.layers.utils import ClipAlpha, NonNeg, NonPos
from decomon.models.utils import get_input_dim


class BackwardDense(BackwardLayer):
    """Backward  LiRPA of Dense"""

    def __init__(
        self,
        layer: Layer,
        input_dim: int = -1,
        rec: int = 1,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            layer=layer,
            rec=rec,
            mode=mode,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            **kwargs,
        )
        self.use_bias = self.layer.use_bias
        if not isinstance(self.layer, DecomonLayer):
            if input_dim < 0:
                input_dim = get_input_dim(self.layer)
            self.layer = to_decomon(
                layer=layer,
                input_dim=input_dim,
                dc_decomp=False,
                perturbation_domain=self.perturbation_domain,
                finetune=False,
                ibp=get_ibp(self.mode),
                affine=get_affine(self.mode),
                shared=True,
                fast=False,
            )
        self.frozen_weights = False

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        if len(inputs) == 0:
            inputs = self.layer.input

        weights = self.kernel
        y = inputs[-1]
        z_value = K.cast(0.0, self.dtype)
        w_u_out, w_l_out = [weights[None] + z_value * K.expand_dims(y, -1)] * 2
        if self.use_bias:
            bias = self.bias
            b_u_out, b_l_out = [bias[None] + z_value * w_u_out[:, 0]] * 2
        else:
            b_u_out, b_l_out = [z_value * w_u_out[:, 0]] * 2
        return [w_u_out, b_u_out, w_l_out, b_l_out]

    def build(self, input_shape: List[tf.TensorShape]) -> None:
        """
        Args:
            input_shape: list of input shape

        Returns:

        """
        if self.layer.kernel is None:
            raise RuntimeError("self.layer.kernel cannot be None when calling self.build()")
        self.kernel = self.layer.kernel
        self._trainable_weights = [self.kernel]
        if self.use_bias:
            if self.layer.bias is None:
                raise RuntimeError("self.layer.bias cannot be None when calling self.build()")
            self.bias = self.layer.bias
            self._trainable_weights.append(self.bias)
        self.built = True

    def freeze_weights(self) -> None:
        if not self.frozen_weights:
            self._trainable_weights = []
            if getattr(self.layer, "freeze_weights"):
                self.layer.freeze_weights()
            self.frozen_weights = True

    def unfreeze_weights(self) -> None:
        if self.frozen_weights:
            if getattr(self.layer, "unfreeze_weights"):
                self.layer.unfreeze_weights()
            self.frozen_weights = False


class BackwardConv2D(BackwardLayer):
    """Backward  LiRPA of Conv2D"""

    def __init__(
        self,
        layer: Layer,
        input_dim: int = -1,
        rec: int = 1,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            layer=layer,
            rec=rec,
            mode=mode,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            **kwargs,
        )
        if not isinstance(self.layer, DecomonLayer):
            if input_dim < 0:
                input_dim = get_input_dim(self.layer)
            self.layer = to_decomon(
                layer=layer,
                input_dim=input_dim,
                dc_decomp=False,
                perturbation_domain=self.perturbation_domain,
                finetune=False,
                ibp=get_ibp(self.mode),
                affine=get_affine(self.mode),
                shared=True,
                fast=False,
            )
        self.frozen_weights = False

    def get_affine_components(self, inputs: List[tf.Tensor]) -> Tuple[tf.Tensor, tf.Tensor]:
        """Express the implicit affine matrix of the convolution layer.

        Conv is a linear operator but its affine component is implicit
        we use im2col and extract_patches to express the affine matrix
        Note that this matrix is Toeplitz

        Args:
            inputs: list of input tensors
        Returns:
            the affine operators W, b : conv(inputs)= W.inputs + b
        """

        w_out_u_ = get_toeplitz(self.layer, True)
        output_shape = self.layer.get_output_shape_at(0)
        if isinstance(output_shape, list):
            output_shape = output_shape[-1]
        output_shape = output_shape[1:]
        if self.layer.data_format == "channels_last":
            b_out_u_ = K.reshape(K.zeros(output_shape, dtype=self.layer.dtype), (-1, output_shape[-1]))
        else:
            b_out_u_ = K.permute_dimensions(
                K.reshape(K.zeros(output_shape, dtype=self.layer.dtype), (-1, output_shape[0])), (1, 0)
            )

        if self.layer.use_bias:
            bias_ = K.cast(self.layer.bias, self.layer.dtype)
            b_out_u_ = b_out_u_ + bias_[None]
        b_out_u_ = K.flatten(b_out_u_)

        z_value = K.cast(0.0, self.dtype)
        y_ = inputs[-1]
        shape = np.prod(y_.shape[1:])
        y_flatten = K.reshape(z_value * y_, (-1, np.prod(shape)))  # (None, n_in)
        w_out_ = K.sum(y_flatten, -1)[:, None, None] + w_out_u_
        b_out_ = K.sum(y_flatten, -1)[:, None] + b_out_u_

        return w_out_, b_out_

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        weight_, bias_ = self.get_affine_components(inputs)
        return [weight_, bias_] * 2

    def freeze_weights(self) -> None:
        if not self.frozen_weights:
            self._trainable_weights = []
            if getattr(self.layer, "freeze_weights"):
                self.layer.freeze_weights()
            self.frozen_weights = True

    def unfreeze_weights(self) -> None:
        if self.frozen_weights:
            if getattr(self.layer, "unfreeze_weights"):
                self.layer.unfreeze_weights()
            self.frozen_weights = False


class BackwardActivation(BackwardLayer):
    def __init__(
        self,
        layer: Layer,
        slope: Union[str, Slope] = Slope.V_SLOPE,
        finetune: bool = False,
        rec: int = 1,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            layer=layer,
            rec=rec,
            mode=mode,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            **kwargs,
        )
        self.activation = get(layer.get_config()["activation"])
        self.activation_name = layer.get_config()["activation"]
        self.slope = Slope(slope)
        self.finetune = finetune
        self.finetune_param: List[tf.Variable] = []
        if self.finetune:
            self.frozen_alpha = False
        self.grid_finetune: List[tf.Variable] = []
        self.frozen_grid = False

    def get_config(self) -> Dict[str, Any]:
        config = super().get_config()
        config.update(
            {
                "slope": self.slope,
                "finetune": self.finetune,
            }
        )
        return config

    def build(self, input_shape: List[tf.TensorShape]) -> None:
        """
        Args:
            input_shape: list of input shape

        Returns:

        """
        input_dim = np.prod(input_shape[-1][1:])

        if self.finetune and self.activation_name != "linear":
            if isinstance(self.perturbation_domain, GridDomain):
                if self.activation_name[:4] == "relu":
                    self.alpha_b_l = self.add_weight(
                        shape=(
                            3,
                            input_dim,
                        ),
                        initializer="ones",
                        name="alpha_l_b_0",
                        regularizer=None,
                        constraint=ClipAlpha(),
                    )
                    alpha_b_l = np.zeros((3, input_dim))
                    alpha_b_l[0] = 1
                    K.set_value(self.alpha_b_l, alpha_b_l)
                    self.finetune_param.append(self.alpha_b_l)

            else:
                self.alpha_b_l = self.add_weight(
                    shape=(input_dim,),
                    initializer="ones",
                    name="alpha_l_b",
                    regularizer=None,
                    constraint=ClipAlpha(),
                )
                alpha_b_l = np.ones((input_dim,))
                # alpha_b_l[0] = 1
                K.set_value(self.alpha_b_l, alpha_b_l)

                if self.activation_name[:4] != "relu":
                    self.alpha_b_u = self.add_weight(
                        shape=(input_dim,),
                        initializer="ones",
                        name="alpha_u_b",
                        regularizer=None,
                        constraint=ClipAlpha(),
                    )
                    self.finetune_param.append(self.alpha_b_u)

                self.finetune_param.append(self.alpha_b_l)
            if len(self.finetune_param) == 1:
                self.finetune_param = self.finetune_param[0]

        # grid domain
        if self.activation_name[:4] == "relu":
            if (
                isinstance(self.perturbation_domain, GridDomain)
                and self.perturbation_domain.opt_option == Option.lagrangian
                and self.mode != ForwardMode.IBP
            ):
                finetune_grid_pos = self.add_weight(
                    shape=(input_dim,),
                    initializer="zeros",
                    name="lambda_grid_neg",
                    regularizer=None,
                    constraint=NonNeg(),
                )

                finetune_grid_neg = self.add_weight(
                    shape=(input_dim,),
                    initializer="zeros",
                    name="lambda_grid_pos",
                    regularizer=None,
                    constraint=NonPos(),
                )

                self.grid_finetune = [finetune_grid_neg, finetune_grid_pos]

        if (
            isinstance(self.perturbation_domain, GridDomain)
            and self.perturbation_domain.opt_option == Option.milp
            and self.mode != ForwardMode.IBP
        ):
            finetune_grid_A = self.add_weight(
                shape=(input_dim,),
                initializer="zeros",
                name=f"A_{self.layer.name}_{self.rec}",
                regularizer=None,
                trainable=False,
            )  # constraint=NonPos()
            finetune_grid_B = self.add_weight(
                shape=(input_dim,),
                initializer="zeros",
                name=f"B_{self.layer.name}_{self.rec}",
                regularizer=None,
                trainable=False,
            )  # constraint=NonNeg()

            self.grid_finetune = [finetune_grid_A, finetune_grid_B]

        self.built = True

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        # infer the output dimension
        if self.activation_name != "linear":
            if self.finetune:
                w_u_out, b_u_out, w_l_out, b_l_out = self.activation(
                    inputs,
                    perturbation_domain=self.perturbation_domain,
                    slope=self.slope,
                    mode=self.mode,
                    finetune=self.finetune_param,
                    finetune_grid=self.grid_finetune,
                )
            else:
                w_u_out, b_u_out, w_l_out, b_l_out = self.activation(
                    inputs,
                    perturbation_domain=self.perturbation_domain,
                    slope=self.slope,
                    mode=self.mode,
                    finetune_grid=self.grid_finetune,
                )
        else:
            w_u_out, b_u_out, w_l_out, b_l_out = get_identity_lirpa(inputs)

        if len(w_u_out.shape) == 2:
            w_u_out = tf.linalg.diag(w_u_out)
        if len(w_l_out.shape) == 2:
            w_l_out = tf.linalg.diag(w_l_out)

        return [w_u_out, b_u_out, w_l_out, b_l_out]

    def freeze_alpha(self) -> None:
        if not self.frozen_alpha:
            if self.finetune and self.mode in [ForwardMode.AFFINE, ForwardMode.HYBRID]:
                if len(self.grid_finetune):
                    self._trainable_weights = self._trainable_weights[:2]
                else:
                    self._trainable_weights = []
                self.frozen_alpha = True

    def unfreeze_alpha(self) -> None:
        if self.frozen_alpha:
            if self.finetune and self.mode in [ForwardMode.AFFINE, ForwardMode.HYBRID]:
                if self.activation_name != "linear":
                    if self.activation_name[:4] != "relu":
                        self._trainable_weights += [self.alpha_b_u, self.alpha_b_l]
                    else:
                        self._trainable_weights += [self.alpha_b_l]
            self.frozen_alpha = False

    def freeze_grid(self) -> None:
        if len(self.grid_finetune) and not self.frozen_grid:
            self._trainable_weights = self._trainable_weights[2:]
            self.frozen_grid = True

    def unfreeze_grid(self) -> None:
        if len(self.grid_finetune) and self.frozen_grid:
            self._trainable_weights = self.grid_finetune + self._trainable_weights
            self.frozen_grid = False


class BackwardFlatten(BackwardLayer):
    """Backward  LiRPA of Flatten"""

    def __init__(
        self,
        layer: Layer,
        rec: int = 1,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            layer=layer,
            rec=rec,
            mode=mode,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            **kwargs,
        )

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        return get_identity_lirpa(inputs)


class BackwardReshape(BackwardLayer):
    """Backward  LiRPA of Reshape"""

    def __init__(
        self,
        layer: Layer,
        rec: int = 1,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            layer=layer,
            rec=rec,
            mode=mode,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            **kwargs,
        )

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        return get_identity_lirpa(inputs)


class BackwardPermute(BackwardLayer):
    """Backward LiRPA of Permute"""

    def __init__(
        self,
        layer: Layer,
        rec: int = 1,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            layer=layer,
            rec=rec,
            mode=mode,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            **kwargs,
        )
        self.dims = layer.dims
        self.op = layer.call

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        w_u_out, b_u_out, w_l_out, b_l_out = get_identity_lirpa(inputs)

        # w_u_out (None, n_in, n_out)
        y = inputs[-1]
        n_dim = w_u_out.shape[1]
        n_out = w_u_out.shape[-1]
        shape = list(y.shape[1:])

        w_u_out = K.reshape(w_u_out, [-1] + shape + [n_out])
        w_l_out = K.reshape(w_l_out, [-1] + shape + [n_out])

        dims = [0] + list(self.dims) + [len(y.shape)]
        dims = list(np.argsort(dims))
        w_u_out = K.reshape(K.permute_dimensions(w_u_out, dims), (-1, n_dim, n_out))
        w_l_out = K.reshape(K.permute_dimensions(w_l_out, dims), (-1, n_dim, n_out))

        return [w_u_out, b_u_out, w_l_out, b_l_out]


class BackwardDropout(BackwardLayer):
    """Backward  LiRPA of Dropout"""

    def __init__(
        self,
        layer: Layer,
        rec: int = 1,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            layer=layer,
            rec=rec,
            mode=mode,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            **kwargs,
        )

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        return get_identity_lirpa(inputs)


class BackwardBatchNormalization(BackwardLayer):
    """Backward  LiRPA of Batch Normalization"""

    def __init__(
        self,
        layer: Layer,
        rec: int = 1,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            layer=layer,
            rec=rec,
            mode=mode,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            **kwargs,
        )

        if not isinstance(layer, DecomonBatchNormalization):
            raise NotImplementedError()
        self.axis = self.layer.axis
        self.op_flat = Flatten()

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        y = inputs[-1]
        w_u_out, b_u_out, w_l_out, b_l_out = get_identity_lirpa(inputs)

        n_dim = y.shape[1:]
        n_out = w_u_out.shape[-1]
        # reshape
        w_u_out = K.reshape(w_u_out, [-1, 1] + list(n_dim) + [n_out])
        w_l_out = K.reshape(w_l_out, [-1, 1] + list(n_dim) + [n_out])

        n_dim = len(y.shape)
        shape = [1] * n_dim
        for i, ax in enumerate(self.axis):
            shape[ax] = self.layer.moving_mean.shape[i]

        gamma = K.reshape(self.layer.gamma + 0.0, shape)
        beta = K.reshape(self.layer.beta + 0.0, shape)
        moving_mean = K.reshape(self.layer.moving_mean + 0.0, shape)
        moving_variance = K.reshape(self.layer.moving_variance + 0.0, shape)

        w = gamma / K.sqrt(moving_variance + self.layer.epsilon)
        b = beta - w * moving_mean

        # flatten w_, b_
        w = K.expand_dims(K.expand_dims(w, -1), 1)
        b = K.expand_dims(K.expand_dims(b, -1), 1)

        n_dim = np.prod(y.shape[1:])
        w_u_b = K.reshape(w_u_out * w, (-1, n_dim, n_out))
        w_l_b = K.reshape(w_l_out * w, (-1, n_dim, n_out))
        axis = [i for i in range(2, len(b.shape) - 1)]
        b_u_b = K.sum(w_u_out * b, axis) + b_u_out
        b_l_b = K.sum(w_l_out * b, axis) + b_l_out

        return [w_u_b, b_u_b, w_l_b, b_l_b]


class BackwardInputLayer(BackwardLayer):
    def __init__(
        self,
        layer: Layer,
        rec: int = 1,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            layer=layer,
            rec=rec,
            mode=mode,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            **kwargs,
        )

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        return get_identity_lirpa(inputs)



import numpy as np
import pytest
import tensorflow.keras.backend as K
from numpy.testing import assert_almost_equal

from decomon.backward_layers.activations import backward_relu, backward_softsign
from decomon.core import ForwardMode, Slope


@pytest.mark.parametrize(
    "activation_func",
    [
        backward_relu,
        backward_softsign,
    ],
)
def test_activation_backward_1D_box(n, mode, floatx, decimal, activation_func, helpers):
    dc_decomp = False

    #  tensor inputs
    inputs = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)

    # numpy inputs
    inputs_ = helpers.get_standard_values_1d_box(n, dc_decomp=dc_decomp)

    # backward outputs
    outputs = activation_func(inputs_for_mode, mode=mode)
    f_func = K.function(inputs, outputs)
    outputs_ = f_func(inputs_)

    # check bounds consistency
    helpers.assert_backward_layer_output_properties_box_linear(
        full_inputs=inputs_,
        backward_outputs=outputs_,
        decimal=decimal,
    )


@pytest.mark.parametrize(
    "n",
    [
        0,
        1,
        2,
    ],
)
def test_activation_backward_1D_box_slope(n, slope, helpers):
    mode = ForwardMode.AFFINE
    activation_func = backward_relu
    dc_decomp = False

    #  tensor inputs
    inputs = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)

    # numpy inputs
    inputs_ = helpers.get_standard_values_1d_box(n, dc_decomp=dc_decomp)
    x_, y_, z_, u_c_, W_u_, B_u_, l_c_, W_l_, B_l_ = inputs_

    # backward outputs
    outputs = activation_func(inputs_for_mode, mode=mode)
    f_func = K.function(inputs, outputs)
    outputs_ = f_func(inputs_)

    # backward recomposition
    w_u_, b_u_, w_l_, b_l_ = outputs_
    w_u_b = np.sum(np.maximum(w_u_, 0) * W_u_ + np.minimum(w_u_, 0) * W_l_, 1)[:, :, None]
    b_u_b = b_u_ + np.sum(np.maximum(w_u_, 0) * B_u_[:, :, None], 1) + np.sum(np.minimum(w_u_, 0) * B_l_[:, :, None], 1)
    w_l_b = np.sum(np.maximum(w_l_, 0) * W_l_ + np.minimum(w_l_, 0) * W_u_, 1)[:, :, None]
    b_l_b = b_l_ + np.sum(np.maximum(w_l_, 0) * B_l_[:, :, None], 1) + np.sum(np.minimum(w_l_, 0) * B_u_[:, :, None], 1)

    # check bounds according to slope
    slope = Slope(slope)
    if n == 0:
        assert_almost_equal(w_u_b, np.zeros(w_u_b.shape))
        assert_almost_equal(b_u_b, np.zeros(b_u_b.shape))
        assert_almost_equal(w_l_b, np.zeros(w_l_b.shape))
        assert_almost_equal(b_l_b, np.zeros(b_l_b.shape))
    elif n == 1:
        assert_almost_equal(w_u_b, len(w_u_b) * np.ones(w_u_b.shape))  # * len ??
        assert_almost_equal(b_u_b, np.zeros(b_u_b.shape))
        assert_almost_equal(w_l_b, len(w_l_b) * np.ones(w_l_b.shape))
        assert_almost_equal(b_l_b, np.zeros(b_l_b.shape))
    elif n == 2:
        assert_almost_equal(w_u_b, 0.5 * len(w_u_b) * np.ones(w_u_b.shape))
        assert_almost_equal(b_u_b, 0.5 * np.ones(b_u_b.shape))
        if slope == Slope.O_SLOPE:
            assert_almost_equal(w_l_b, np.zeros(w_l_b.shape))  # ??
            assert_almost_equal(b_l_b, np.zeros(b_l_b.shape))
        elif slope == Slope.Z_SLOPE:
            assert_almost_equal(w_l_b, np.zeros(w_l_b.shape))
            assert_almost_equal(b_l_b, np.zeros(b_l_b.shape))
        elif slope == Slope.V_SLOPE:
            assert_almost_equal(w_l_b, np.zeros(w_l_b.shape))
            assert_almost_equal(b_l_b, np.zeros(b_l_b.shape))
        elif slope == Slope.S_SLOPE:
            assert_almost_equal(w_l_b, np.zeros(w_l_b.shape))  # ??
            assert_almost_equal(b_l_b, np.zeros(b_l_b.shape))
        elif slope == Slope.A_SLOPE:
            assert_almost_equal(b_l_b, np.zeros(b_l_b.shape))
            assert_almost_equal(w_l_b, np.zeros(w_l_b.shape))



"""Extract python code blocks from a markdown file to generate a python script

The goal is to be able to test python code examples shown in a markdown file.

Usage:
    md2py.py <md_inputfile> <py_outputfile>

Arguments:
    <md_inputfile>   path of the markdownfile from which extract python code blocks
    <py_outputfile>  path of the resulting python script

"""

from __future__ import annotations

import commonmark
from docopt import docopt

PYTHON_FLAVORS = ["python", "py3", "python3"]


def extract_pythoncode_from_markdown(markdown_inputfile: str) -> list[str]:
    with open(markdown_inputfile, "rt") as fp:
        doc = fp.read()
    python_nodes = []
    parser = commonmark.Parser()
    ast = parser.parse(doc)
    walker = ast.walker()
    for node, entering in walker:
        if node.t == "code_block" and node.is_fenced and node.info in PYTHON_FLAVORS:
            python_nodes.append(node)
    return [node.literal for node in python_nodes]


def wrap_pythoncode_in_testfunction(pythoncodes: list[str]) -> str:
    docstring = '"""Python code auto-generated by scripts/md2py.py from code blocks within a markdown file."""\n'
    testfunction_header_pattern = "def test_code_within_markdown_{}():"
    tabs = "    "
    testfunctions: list[str] = []
    for i, pythoncode in enumerate(pythoncodes):
        testfunctions.append(
            "\n".join(
                [testfunction_header_pattern.format(i)]
                + [tabs + line if line else line for line in pythoncode.split("\n")]
            )
        )
    return "\n\n".join([docstring] + testfunctions)


if __name__ == "__main__":
    #  arguments
    arguments = docopt(__doc__)
    markdown_inputfile = arguments["<md_inputfile>"]
    python_outputfile = arguments["<py_outputfile>"]

    # extract code lines
    pythoncodes = extract_pythoncode_from_markdown(markdown_inputfile=markdown_inputfile)

    # wrap them in a test function
    python_string = wrap_pythoncode_in_testfunction(pythoncodes=pythoncodes)

    with open(python_outputfile, "wt") as fp:
        fp.write(python_string)
